@@ -20,6 +20,7 @@ from typing import (
     Union,
     Sequence,
     Callable,
+    Mapping,
 )
 from pathlib import Path
 
@@ -54,26 +55,69 @@ from ridepy.extras.io import (
 logger = logging.getLogger(__name__)
 
 
-def make_file_path(sim_id: str, directory: Path, suffix: str):
+def make_file_path(sim_id: str, directory: Path, suffix: str) -> Path:
     return directory / f"{sim_id}{suffix}"
 
 
-def get_params(directory: Path, sim_id: str, param_path_suffix: str = "_params.json"):
+def get_params(
+    directory: Path, sim_id: str, param_path_suffix: str = "_params.json"
+) -> dict:
     return read_params_json(make_file_path(sim_id, directory, param_path_suffix))
 
 
 def perform_single_analysis(
     sim_id: str,
     data_dir: Path,
+    tasks_if_not_existent: Optional[Iterable[str]] = None,
+    tasks_if_existent: Optional[Iterable[str]] = None,
     param_path_suffix: str = "_params.json",
     event_path_suffix: str = ".jsonl",
-    update_existing: bool = False,
-    compute_system_quantities: bool = True,
-    compute_vehicle_quantities: bool = True,
     stops_path_suffix: str = "_stops.pq",
     requests_path_suffix: str = "_requests.pq",
     vehicle_quantities_path_suffix: str = "_vehicle_quantities.pq",
-) -> tuple[str, dict]:
+) -> tuple[str, dict[str, Any]]:
+    """
+    Compute stops, requests, vehicle quantities, and system quantities from
+    simulation events.
+
+    Parameters
+    ----------
+    sim_id
+        Simulation ID
+    data_dir
+        Directory from which to read the events and to which to write the resulting
+        parquet output files
+    tasks_if_not_existent
+        Collection of tasks to perform if the respective output files do not exist.
+        Valid entries are 'stops', 'requests', 'vehicle_quantities', and
+        'system_quantities'. If None, everything is computed if not existent.
+    tasks_if_existent
+        Collection of tasks to perform even if the respective output files do exist,
+        i.e. tasks to recompute. Valid entries are 'stops', 'requests',
+        'vehicle_quantities', and 'system_quantities'. If None, nothing is recomputed.
+    param_path_suffix
+        Appending this suffix to the simulation ID yields the path to the JSON parameter
+        file.
+    event_path_suffix
+        Appending this suffix to the simulation ID yields the path to the JSONL
+        event file.
+    stops_path_suffix
+        Appending this suffix to the simulation ID yields the path to the parquet
+        stops file.
+    requests_path_suffix
+        Appending this suffix to the simulation ID yields the path to the parquet
+        requests file.
+    vehicle_quantities_path_suffix
+        Appending this suffix to the simulation ID yields the path to the \
+        parquet vehicle quantities file.
+
+    Returns
+    -------
+    sim_id
+        Simulation ID
+    system_quantities
+        Dictionary containing the computed system quantities as entries
+    """
     stops_path = make_file_path(sim_id, data_dir, stops_path_suffix)
     requests_path = make_file_path(sim_id, data_dir, requests_path_suffix)
     vehicle_quantities_path = make_file_path(
@@ -82,16 +126,27 @@ def perform_single_analysis(
 
     tasks = set()
 
-    if update_existing:
-        tasks |= {"stops", "requests", "vehicle_quantities", "system_quantities"}
-    if not stops_path.exists():
+    tasks_if_not_existent = tasks_if_not_existent or {
+        "stops",
+        "requests",
+        "vehicle_quantities",
+        "system_quantities",
+    }
+    tasks_if_existent = tasks_if_existent or set()
+
+    if "stops" in tasks_if_not_existent and not stops_path.exists():
         tasks.add("stops")
-    if not requests_path.exists():
+    if "requests" in tasks_if_not_existent and not requests_path.exists():
         tasks.add("requests")
-    if compute_system_quantities:
-        tasks.add("system_quantities")
-    if compute_vehicle_quantities and not vehicle_quantities_path.exists():
+    if (
+        "vehicle_quantities" in tasks_if_not_existent
+        and not vehicle_quantities_path.exists()
+    ):
         tasks.add("vehicle_quantities")
+    if "system_quantities" in tasks_if_not_existent:
+        tasks.add("system_quantities")
+
+    tasks |= tasks_if_existent
 
     system_quantities = {}
     if tasks:
@@ -240,25 +295,21 @@ def perform_single_simulation(
     if (
         params["general"].get("n_vehicles") is not None
         and params["general"].get("initial_location") is not None
+        and params["general"].get("initial_locations") is None
     ):
         initial_locations = {
             vehicle_id: params["general"]["initial_location"]
             for vehicle_id in range(params["general"]["n_vehicles"])
         }
-
-        if params["general"].get("initial_locations") is not None:
-            warnings.warn(f"disregarding {params['general']['initial_locations']=}")
-
-    elif params["general"].get("initial_locations") is not None:
+    elif (
+        params["general"].get("n_vehicles") is None
+        and params["general"].get("initial_location") is None
+        and params["general"].get("initial_locations") is not None
+    ):
         initial_locations = params["general"]["initial_locations"]
-
-        if params["general"].get("initial_location") is not None:
-            warnings.warn(f"disregarding {params['general']['initial_location']=}")
-        if params["general"].get("n_vehicles") is not None:
-            warnings.warn(f"disregarding {params['general']['n_vehicles']=}")
     else:
         raise ValueError(
-            "must either specify 'n_vehicles' and 'initial_location' or 'initial_locations'"
+            "Must *either* specify `n_vehicles` *and* `initial_location` *or* `initial_locations`"
         )
 
     fs = params["general"]["fleet_state_cls"](
@@ -274,12 +325,18 @@ def perform_single_simulation(
         print(f"Simulating run on process {os.getpid()} @ \n{params!r}\n")
 
     if not dry_run:
-        if params["general"]["n_reqs"] is not None:
+        if (
+            params["general"]["n_reqs"] is not None
+            and params["general"]["t_cutoff"] is None
+        ):
             simulation = fs.simulate(it.islice(rg, params["general"]["n_reqs"]))
-        elif params["general"]["t_cutoff"] is not None:
+        elif (
+            params["general"]["n_reqs"] is None
+            and params["general"]["t_cutoff"] is not None
+        ):
             simulation = fs.simulate(rg, t_cutoff=params["general"]["t_cutoff"])
         else:
-            raise ValueError("must either specify n_reqs or t_cutoff")
+            raise ValueError("Must *either* specify `n_reqs` *or* `t_cutoff`")
 
         while chunk := list(it.islice(simulation, jsonl_chunksize)):
             save_events_json(jsonl_path=event_path, events=chunk)
@@ -373,10 +430,15 @@ class SimulationSet:
 
         .. code-block:: python
 
-            >>> SimulationSet._two_level_dict_update(
-            ...     {"a": {"b": 5, 6: 8}}, {"a": {6: "fooo", 8: "baaar"}, "baz": {"6": 6}}
-            ... )
-            {'a': {'b': 5, 6: 'fooo', 8: 'baaar'}, 'baz': {'6': 6}}
+            SimulationSet._two_level_dict_update(
+                {"a": {"b": 5, 6: 8}}, {"a": {6: "fooo", 8: "baaar"}, "baz": {"6": 6}}
+            )
+
+        yields
+
+        .. code-block:: python
+
+            {"a": {"b": 5, 6: "fooo", 8: "baaar"}, "baz": {"6": 6}}
 
 
         Parameters
@@ -511,9 +573,6 @@ class SimulationSet:
             transportation_request_cls = TransportationRequest
             vehicle_state_cls = VehicleState
 
-        fleet_state_cls = SlowSimpleFleetState
-        request_generator_cls = RandomRequestGenerator
-
         self.default_base_params = dict(
             general=dict(
                 n_reqs=100,
@@ -521,14 +580,16 @@ class SimulationSet:
                 space=space_obj,
                 n_vehicles=10,
                 initial_location=(0, 0),
+                initial_locations=None,
                 seat_capacity=8,
                 transportation_request_cls=transportation_request_cls,
                 vehicle_state_cls=vehicle_state_cls,
-                fleet_state_cls=fleet_state_cls,
+                fleet_state_cls=SlowSimpleFleetState,
             ),
             dispatcher=dict(dispatcher_cls=dispatcher),
             request_generator=dict(
-                request_generator_cls=request_generator_cls,
+                request_generator_cls=RandomRequestGenerator,
+                rate=1,
             ),
         )
 
@@ -568,7 +629,7 @@ class SimulationSet:
 
         self._simulation_ids = None
 
-        self.system_quantities_path = None
+        self._system_quantities_path = None
 
     @property
     def simulation_ids(self) -> list[str]:
@@ -725,27 +786,80 @@ class SimulationSet:
 
     def run_analytics(
         self,
-        update_existing: bool = False,
+        only_stops_and_requests: bool = False,
+        update_existing: Union[bool, list[str]] = False,
+        check_for_changes: bool = True,
         stops_path_suffix: str = "_stops.pq",
         requests_path_suffix: str = "_requests.pq",
-        only_stops_and_requests: bool = False,  # only compute stops and requests
         vehicle_quantities_path_suffix: str = "_vehicle_quantities.pq",
         system_quantities_filename: str = "system_quantities.pq",
-    ):
-        self.system_quantities_path = self.data_dir / system_quantities_filename
+    ) -> None:
+        """
+        Compute analytics from simulation events and store them to disk
+        in parquet format.
+
+        Parameters
+        ----------
+        only_stops_and_requests
+            Only compute stops and requests, not vehicle and system quantities.
+        update_existing
+            Recompute existing outputs. If a list is given, only recompute
+            the list entries. Valid list items are 'system_quantities',
+            'vehicle_quantities', 'stops', and 'requests'.
+        check_for_changes
+            If True, only update system quantities if simulation ids have changed.
+            If False, do update system quantities in any case.
+        stops_path_suffix
+            Appending this suffix to the simulation ID yields the path to the parquet
+            stops file.
+        requests_path_suffix
+            Appending this suffix to the simulation ID yields the path to the parquet
+            requests file.
+        vehicle_quantities_path_suffix
+            Appending this suffix to the simulation ID yields the path to the parquet
+            vehicle quantities file.
+        system_quantities_filename
+            Filename of the parquet file to store the system quantities in.
+        """
+        self._system_quantities_path = self.data_dir / system_quantities_filename
 
         if not self.simulation_ids:
             warnings.warn(
                 "no simulations have been run (simulation_ids empty)", UserWarning
             )
         else:
-            if only_stops_and_requests:
-                compute_system_quantities = compute_vehicle_quantities = False
+
+            # In any case, stops and requests have to be computed if they don't exist
+            tasks_if_not_existent = {"stops", "requests"}
+
+            if not only_stops_and_requests:
+                # Additionally, we are computing vehicle and system quantities, now,
+                # if they don't exist
+                tasks_if_not_existent |= {"vehicle_quantities", "system_quantities"}
+
+            if isinstance(update_existing, Iterable):
+                # If we have been handed a list of tasks to update, we only update these,
+                # not recomputing tasks that should not have be computed in the first place
+                tasks_if_existent = set(update_existing) & tasks_if_not_existent
+            elif update_existing == True:
+                tasks_if_existent = tasks_if_not_existent
+            elif update_existing == False:
+                tasks_if_existent = set()
             else:
-                compute_vehicle_quantities = True
-                compute_system_quantities = (
-                    not self.system_quantities_path.exists() or update_existing
-                )
+                raise ValueError(f"Got invalid value for {update_existing=}")
+
+            if self._system_quantities_path.exists():
+                tasks_if_not_existent -= {"system_quantities"}
+                if check_for_changes:
+                    # Currently, we only check for changes in the sense that new
+                    # simulations with new ids have been performed and are thus missing from
+                    # the system quantities output. For vehicle quantities, stops, and requests,
+                    # check_for_changes does not apply, as we compute these in any case, should
+                    # they be missing.
+                    sqdf = pd.read_parquet(self._system_quantities_path)
+                    if set(sqdf.index) == set(self.simulation_ids):
+                        tasks_if_existent -= {"system_quantities"}
+                    del sqdf
 
             with loky.get_reusable_executor(max_workers=self.max_workers) as executor:
                 sim_ids, system_quantities = zip(
@@ -754,9 +868,8 @@ class SimulationSet:
                             ft.partial(
                                 perform_single_analysis,
                                 data_dir=self.data_dir,
-                                update_existing=update_existing,
-                                compute_system_quantities=compute_system_quantities,
-                                compute_vehicle_quantities=compute_vehicle_quantities,
+                                tasks_if_not_existent=tasks_if_not_existent,
+                                tasks_if_existent=tasks_if_existent,
                                 param_path_suffix=self._param_path_suffix,
                                 event_path_suffix=self._event_path_suffix,
                                 stops_path_suffix=stops_path_suffix,
@@ -775,7 +888,83 @@ class SimulationSet:
                     UserWarning,
                 )
 
-            if compute_system_quantities:
+            if "system_quantities" in tasks_if_not_existent | tasks_if_existent:
                 system_quantities_df = pd.DataFrame(system_quantities, index=sim_ids)
                 system_quantities_df.rename_axis("simulation_id", inplace=True)
-                system_quantities_df.to_parquet(self.system_quantities_path)
+                system_quantities_df.to_parquet(self._system_quantities_path)
+
+    @property
+    def param_path_suffix(self) -> str:
+        """
+        Get the parameter file suffix.
+        """
+        return self._param_path_suffix
+
+    @property
+    def event_path_suffix(self) -> str:
+        """
+        Get the event file suffix.
+        """
+        return self._event_path_suffix
+
+    @property
+    def system_quantities_path(self) -> Path:
+        """
+        Get path to the parquet file containing the system quantities.
+
+        Returns
+        -------
+        system_quantities_path
+        """
+        if self._system_quantities_path is not None:
+            return self._system_quantities_path
+        else:
+            raise AttributeError("No system quantities path set.")
+
+    def get_system_quantities(
+        self, extra_params: Optional[Mapping] = None
+    ) -> pd.DataFrame:
+        """
+        Return the system quantities, if computed already using `run_analytics`.
+
+        Optionally, join the system quantities dataframe with additional arbitrary
+        parameters used in the simulation.
+
+        Parameters
+        ----------
+        extra_params:
+            Dictionary specifying the desired parameters. The keys are the resulting column
+            names in the returned dataframe, the values are the (nested) keys in the
+            `SimulationSet` params dictionary, joined by ``.``.
+
+            Example:
+
+            .. code-block:: python
+
+                extra_params = {"n": "general.n_reqs"}
+
+        Returns
+        -------
+        system_quantities
+            DataFrame containing the system quantities and optional extra parameters
+
+        """
+
+        try:
+            sqdf = pd.read_parquet(self.system_quantities_path)
+        except FileNotFoundError:
+            raise FileNotFoundError(
+                f"System properties file {self.system_quantities_path} not found."
+            )
+
+        if extra_params:
+            for sim_id in sqdf.index:
+                params = get_params(
+                    self.data_dir, sim_id, param_path_suffix=self.param_path_suffix
+                )
+                for param_col, param_key in extra_params.items():
+                    sqdf.loc[sim_id, param_col] = ft.reduce(
+                        op.getitem, param_key.split("."), params
+                    )
+
+        return sqdf
